import 'package:endlich/endlich.dart';
import 'package:riverpod/riverpod.dart';

import 'builder.dart';
import 'common.dart';

abstract class AutoDisposeMachineProviderRef<StateValue, Event>
    implements AutoDisposeProviderRefBase {
  /// Adding a state to the [StateMachine].
  /// The given callback will always execute when the state is entered.
  ///
  /// This method must only be called while the provider is being created.
  void onState<S extends StateValue>(OnEnterState<StateValue, S, Event> cb);

  /// Automatically start the Machine in running status.
  /// It takes the state returned by the providers body as the initial state
  void startAsRunning();
}

class AutoDisposeMachineProviderElement<StateValue, Event>
    extends AutoDisposeProviderElementBase<Machine<StateValue, Event>>
    implements AutoDisposeMachineProviderRef<StateValue, Event> {
  AutoDisposeMachineProviderElement(
      ProviderBase<Machine<StateValue, Event>> provider)
      : super(provider);

  late MachineBuilder<StateValue, Event> _builder =
      MachineBuilder<StateValue, Event>();

  @override
  void onState<S extends StateValue>(OnEnterState<StateValue, S, Event> cb) =>
      _builder.onState(cb);

  @override
  void startAsRunning() => _builder.startAsRunning();
}

class _AutoDisposeMachineProvider<StateValue, Event>
    extends AutoDisposeProviderBase<Machine<StateValue, Event>> {
  _AutoDisposeMachineProvider(this._create, {required String? name})
      : super(name == null ? null : '$name.machine');

  final Create<StateValue, AutoDisposeMachineProviderRef<StateValue, Event>>
      _create;

  @override
  Machine<StateValue, Event> create(
      AutoDisposeMachineProviderRef<StateValue, Event> ref) {
    final ele = ref as AutoDisposeMachineProviderElement<StateValue, Event>;

    final machine = Machine<StateValue, Event>((b) {
      ele._builder = b;
      return _create(ele);
    });
    ref.onDispose(machine.dispose);

    return machine;
  }

  @override
  bool updateShouldNotify(Machine<StateValue, Event> previousState,
      Machine<StateValue, Event> newState) {
    return true;
  }

  @override
  AutoDisposeMachineProviderElement<StateValue, Event> createElement() =>
      AutoDisposeMachineProviderElement<StateValue, Event>(this);

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override MachineProvider.notifier');
}

class _AutoDisposeCustomMachineProvider<StateValue, Event>
    extends AutoDisposeProviderBase<Machine<StateValue, Event>> {
  _AutoDisposeCustomMachineProvider(this._create, {required String? name})
      : super(name == null ? null : '$name.machine');

  final Create<Machine<StateValue, Event>, AutoDisposeProviderRefBase> _create;

  @override
  Machine<StateValue, Event> create(AutoDisposeProviderRefBase ref) {
    final machine = _create(ref);
    ref.onDispose(machine.dispose);
    return machine;
  }

  @override
  bool updateShouldNotify(Machine<StateValue, Event> previousState,
      Machine<StateValue, Event> newState) {
    return true;
  }

  @override
  AutoDisposeProviderElement<Machine<StateValue, Event>> createElement() =>
      AutoDisposeProviderElement<Machine<StateValue, Event>>(this);

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override MachineProvider.notifier');
}

class AutoDisposeStateMachineProvider<StateValue, Event>
    extends AutoDisposeProviderBase<MachineStatus<StateValue, Event>>
    with StateMachineProviderMixin<StateValue, Event> {
  AutoDisposeStateMachineProvider(this._create, {String? name})
      : _createMachine = null,
        super(name);

  AutoDisposeStateMachineProvider.custom(this._createMachine, {String? name})
      : _create = null,
        super(name);

  static const family = AutoDisposeStateMachineProviderFamilyBuilder();

  final Create<StateValue, AutoDisposeMachineProviderRef<StateValue, Event>>?
      _create;

  final Create<Machine<StateValue, Event>, AutoDisposeProviderRefBase>?
      _createMachine;

  @override
  late final AutoDisposeProviderBase<Machine<StateValue, Event>> machine =
      _create != null
          ? _AutoDisposeMachineProvider<StateValue, Event>(_create!, name: name)
          : _AutoDisposeCustomMachineProvider<StateValue, Event>(
              _createMachine!,
              name: name);

  @override
  late final AutoDisposeProviderBase<MachineSend<Event>> send =
      AutoDisposeProvider<MachineSend<Event>>((ref) => ref.watch(machine).send,
          name: (name ?? '') + '.send');

  @override
  MachineStatus<StateValue, Event> create(
      AutoDisposeProviderElementBase<MachineStatus<StateValue, Event>> ref) {
    final machine = ref.watch(this.machine);

    ref.onDispose(machine.addListener(
      (status) {
        ref.state = status;
      },
      fireImmediately: false,
    ));

    return machine.state;
  }

  @override
  bool updateShouldNotify(MachineStatus<StateValue, Event> previousState,
      MachineStatus<StateValue, Event> newState) {
    return true;
  }

  Override overrideWithProvider(
    AutoDisposeStateMachineProvider<StateValue, Event> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: machine, override: provider.machine);
      setup(origin: send, override: provider.send);
      setup(origin: this, override: this);
    });
  }

  @override
  AutoDisposeProviderElementBase<MachineStatus<StateValue, Event>>
      createElement() => AutoDisposeProviderElement(this);
}

class AutoDisposeStateMachineProviderFamily<StateValue, Event, Arg>
    extends Family<MachineStatus<StateValue, Event>, Arg,
        AutoDisposeStateMachineProvider<StateValue, Event>> {
  AutoDisposeStateMachineProviderFamily(this._create, {String? name})
      : _createCustom = null,
        super(name);

  AutoDisposeStateMachineProviderFamily.custom(this._createCustom,
      {String? name})
      : _create = null,
        super(name);

  final FamilyCreate<StateValue,
      AutoDisposeMachineProviderRef<StateValue, Event>, Arg>? _create;

  final FamilyCreate<Machine<StateValue, Event>, AutoDisposeProviderRefBase,
      Arg>? _createCustom;

  @override
  AutoDisposeStateMachineProvider<StateValue, Event> create(
    Arg argument,
  ) {
    AutoDisposeStateMachineProvider<StateValue, Event>? provider;
    if (null != _create) {
      provider = AutoDisposeStateMachineProvider<StateValue, Event>(
        (ref) => _create!(ref, argument),
        name: name,
      );
    } else if (null != _createCustom) {
      provider = AutoDisposeStateMachineProvider<StateValue, Event>.custom(
        (ref) => _createCustom!(ref, argument),
        name: name,
      );
    } else {
      throw StateError('provider must not be null');
    }

    registerProvider(provider.machine, argument);

    return provider;
  }

  Override overrideWithProvider(
    AutoDisposeStateMachineProvider<StateValue, Event> Function(Arg argument)
        override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.machine, override: override(arg).machine);
        setup(origin: provider.send, override: override(arg).send);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.machine, override: provider.machine);
    setup(origin: provider.send, override: provider.send);
  }
}
