import 'package:endlich/endlich.dart';
import 'package:riverpod/riverpod.dart';

import 'builder.dart';
import 'common.dart';

abstract class MachineProviderRef<StateValue, Event>
    implements ProviderRefBase {
  /// Adding a state to the [StateMachine].
  /// The given callback will always execute when the state is entered.
  ///
  /// This method must only be called while the provider is being created.
  void onState<S extends StateValue>(OnEnterState<StateValue, S, Event> cb);

  /// Automatically start the Machine in running status.
  /// It takes the state returned by the providers body as the initial state
  void startAsRunning();
}

class MachineProviderElement<StateValue, Event>
    extends ProviderElementBase<Machine<StateValue, Event>>
    implements MachineProviderRef<StateValue, Event> {
  MachineProviderElement(ProviderBase<Machine<StateValue, Event>> provider)
      : super(provider);

  late MachineBuilder<StateValue, Event> _builder =
      MachineBuilder<StateValue, Event>();

  @override
  void onState<S extends StateValue>(OnEnterState<StateValue, S, Event> cb) =>
      _builder.onState(cb);

  @override
  void startAsRunning() => _builder.startAsRunning();
}

class _MachineProvider<StateValue, Event>
    extends AlwaysAliveProviderBase<Machine<StateValue, Event>> {
  _MachineProvider(this._create, {required String? name})
      : super(name == null ? null : '$name.machine');

  final Create<StateValue, MachineProviderRef<StateValue, Event>> _create;

  @override
  Machine<StateValue, Event> create(MachineProviderRef<StateValue, Event> ref) {
    final ele = ref as MachineProviderElement<StateValue, Event>;

    final machine = Machine<StateValue, Event>((b) {
      ele._builder = b;
      return _create(ele);
    });
    ref.onDispose(machine.dispose);

    return machine;
  }

  @override
  bool updateShouldNotify(Machine<StateValue, Event> previousState,
      Machine<StateValue, Event> newState) {
    return true;
  }

  @override
  MachineProviderElement<StateValue, Event> createElement() =>
      MachineProviderElement<StateValue, Event>(this);

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override MachineProvider.notifier');
}

class _CustomMachineProvider<StateValue, Event>
    extends AlwaysAliveProviderBase<Machine<StateValue, Event>> {
  _CustomMachineProvider(this._create, {required String? name})
      : super(name == null ? null : '$name.machine');

  final Create<Machine<StateValue, Event>, ProviderRefBase> _create;

  @override
  Machine<StateValue, Event> create(ProviderRefBase ref) {
    final machine = _create(ref);
    ref.onDispose(machine.dispose);
    return machine;
  }

  @override
  bool updateShouldNotify(Machine<StateValue, Event> previousState,
      Machine<StateValue, Event> newState) {
    return true;
  }

  @override
  ProviderElement<Machine<StateValue, Event>> createElement() =>
      ProviderElement<Machine<StateValue, Event>>(this);

  @override
  void setupOverride(SetupOverride setup) =>
      throw UnsupportedError('Cannot override MachineProvider.notifier');
}

class StateMachineProvider<StateValue, Event>
    extends AlwaysAliveProviderBase<MachineStatus<StateValue, Event>>
    with StateMachineProviderMixin<StateValue, Event> {
  StateMachineProvider(this._create, {String? name})
      : _createMachine = null,
        super(name);

  StateMachineProvider.custom(this._createMachine, {String? name})
      : _create = null,
        super(name);

  static const family = StateMachineProviderFamilyBuilder();
  static const autoDispose = AutoDisposeStateMachineProviderBuilder();
  static const autoDisposeFamily =
      AutoDisposeStateMachineProviderFamilyBuilder();

  final Create<StateValue, MachineProviderRef<StateValue, Event>>? _create;

  final Create<Machine<StateValue, Event>, ProviderRefBase>? _createMachine;

  @override
  late final AlwaysAliveProviderBase<Machine<StateValue, Event>> machine =
      _create != null
          ? _MachineProvider<StateValue, Event>(_create!, name: name)
          : _CustomMachineProvider<StateValue, Event>(_createMachine!,
              name: name);

  @override
  late final AlwaysAliveProviderBase<MachineSend<Event>> send =
      Provider<MachineSend<Event>>((ref) => ref.watch(machine).send,
          name: (name ?? '') + '.send');

  @override
  MachineStatus<StateValue, Event> create(
      ProviderElementBase<MachineStatus<StateValue, Event>> ref) {
    final machine = ref.watch(this.machine);

    ref.onDispose(machine.addListener(
      (status) {
        ref.state = status;
      },
      fireImmediately: false,
    ));

    return machine.state;
  }

  @override
  bool updateShouldNotify(MachineStatus<StateValue, Event> previousState,
      MachineStatus<StateValue, Event> newState) {
    return true;
  }

  Override overrideWithProvider(
    StateMachineProvider<StateValue, Event> provider,
  ) {
    return ProviderOverride((setup) {
      setup(origin: machine, override: provider.machine);
      setup(origin: send, override: provider.send);
      setup(origin: this, override: this);
    });
  }

  @override
  ProviderElementBase<MachineStatus<StateValue, Event>> createElement() =>
      ProviderElement(this);
}

class StateMachineProviderFamily<StateValue, Event, Arg> extends Family<
    MachineStatus<StateValue, Event>,
    Arg,
    StateMachineProvider<StateValue, Event>> {
  StateMachineProviderFamily(this._create, {String? name})
      : _createCustom = null,
        super(name);
  StateMachineProviderFamily.custom(this._createCustom, {String? name})
      : _create = null,
        super(name);

  final FamilyCreate<StateValue, MachineProviderRef<StateValue, Event>, Arg>?
      _create;

  final FamilyCreate<Machine<StateValue, Event>, ProviderRefBase, Arg>?
      _createCustom;

  @override
  StateMachineProvider<StateValue, Event> create(
    Arg argument,
  ) {
    StateMachineProvider<StateValue, Event>? provider;
    if (null != _create) {
      provider = StateMachineProvider<StateValue, Event>(
        (ref) => _create!(ref, argument),
        name: name,
      );
    } else if (null != _createCustom) {
      provider = StateMachineProvider<StateValue, Event>.custom(
        (ref) => _createCustom!(ref, argument),
        name: name,
      );
    } else {
      throw StateError('provider must not be null');
    }

    registerProvider(provider.machine, argument);

    return provider;
  }

  Override overrideWithProvider(
    StateMachineProvider<StateValue, Event> Function(Arg argument) override,
  ) {
    return FamilyOverride<Arg>(
      this,
      (arg, setup) {
        final provider = call(arg);
        setup(origin: provider.machine, override: override(arg).machine);
        setup(origin: provider.send, override: override(arg).send);
        setup(origin: provider, override: provider);
      },
    );
  }

  @override
  void setupOverride(Arg argument, SetupOverride setup) {
    final provider = call(argument);
    setup(origin: provider, override: provider);
    setup(origin: provider.machine, override: provider.machine);
    setup(origin: provider.send, override: provider.send);
  }
}
