import 'package:endlich/endlich.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:riverpod/riverpod.dart';
import 'package:riverpod_endlich/riverpod_endlich.dart';
import 'package:riverpod_endlich/src/base_provider.dart';
import 'package:test/test.dart';

import 'riverpod_endlich_test.mocks.dart';
part 'riverpod_endlich_test.freezed.dart';

@GenerateMocks([ControllerListenerMock, ListenerMock, Listener, OnStateCb])
void main() {
  ProviderContainer createContainer({
    ProviderContainer? parent,
    List<Override> overrides = const [],
    List<ProviderObserver>? observers,
  }) {
    final container = ProviderContainer(
      parent: parent,
      overrides: overrides,
      observers: observers,
    );
    addTearDown(container.dispose);
    return container;
  }

  Machine<State, Event> fooMachine({
    State? initial,
    bool? autostart,
    void Function(ActiveState<State, _Foo, Event> self)? onFoo,
    void Function(ActiveState<State, _Bar, Event> self)? onBar,
    void Function(ActiveState<State, _Baz, Event> self)? onBaz,
  }) {
    return Machine<State, Event>((b) {
      b.onState<_Foo>((self) {
        if (null != onFoo) onFoo(self);
      });
      b.onState<_Bar>((self) {
        if (null != onBar) onBar(self);
      });
      b.onState<_Baz>((self) {
        if (null != onBaz) onBaz(self);
      });
      if (true == autostart) {
        b.startAsRunning();
      }
      return initial ?? const State.foo();
    });
  }

  StateMachineProvider<State, Event> fooMachineProvider({
    State Function()? initial,
    bool? autostart,
    void Function(ActiveState<State, _Foo, Event> self)? onFoo,
    void Function(ActiveState<State, _Bar, Event> self)? onBar,
    void Function(ActiveState<State, _Baz, Event> self)? onBaz,
    void Function(MachineProviderRef<State, Event> ref)? onRef,
    String? name,
  }) {
    return StateMachineProvider<State, Event>(
      (ref) {
        onRef?.call(ref);
        ref.onState<_Foo>((self) {
          if (null != onFoo) onFoo(self);
        });
        ref.onState<_Bar>((self) {
          if (null != onBar) onBar(self);
        });
        ref.onState<_Baz>((self) {
          if (null != onBaz) onBaz(self);
        });
        if (true == autostart) {
          ref.startAsRunning();
        }
        return initial?.call() ?? const State.foo();
      },
      name: name,
    );
  }

  test('can be refreshed', () async {
    var result = fooMachine(autostart: true);
    final container = createContainer();
    final provider = StateMachineProvider<State, Event>.custom((ref) => result);

    expect(
        container.read(provider),
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.foo()));
    expect(container.read(provider.machine), result);

    result = fooMachine(autostart: false);
    expect(container.refresh(provider), isA<MachineStatusIdle<State, Event>>());

    expect(container.read(provider), isA<MachineStatusIdle<State, Event>>());
    expect(container.read(provider.machine), result);
  });

  group('scope', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final machine = fooMachine(autostart: true);
        final provider =
            StateMachineProvider<State, Event>.custom((ref) => machine);
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider.machine), machine);
        expect(container.read(provider),
            isA<MachineStatusRunning<State, Event>>());
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });

      test('when using provider.overrideWithValue', () async {
        final machine = fooMachine(autostart: true);
        final provider =
            StateMachineProvider<State, Event>.custom((ref) => machine);
        final root = createContainer();
        final machinerOverride = fooMachine(autostart: false);
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithValue(machinerOverride),
        ]);

        expect(container.read(provider.machine), machinerOverride);
        expect(
            container.read(provider), isA<MachineStatusIdle<State, Event>>());
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });

      test('when using provider.overrideWithProvider', () async {
        final machine = fooMachine(autostart: true);
        final provider =
            StateMachineProvider<State, Event>.custom((ref) => machine);
        final root = createContainer();
        final machinerOverride = fooMachine(autostart: false);
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            StateMachineProvider<State, Event>.custom(
                (ref) => machinerOverride),
          ),
        ]);

        expect(container.read(provider.machine), machinerOverride);
        expect(
            container.read(provider), isA<MachineStatusIdle<State, Event>>());
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });
    });
  });

  test('StateNotifierFamily override', () {
    final machine = fooMachine(autostart: true);
    final provider =
        StateMachineProvider.family.custom<State, Event, int>((ref, a) {
      return machine;
    });

    final machine2 = fooMachine(autostart: false);
    final container = createContainer(
      overrides: [
        provider.overrideWithProvider((a) {
          return StateMachineProvider<State, Event>.custom(
            (ref) => machine2,
          );
        }),
      ],
    );
    addTearDown(container.dispose);

    // try to read provider.state before provider and see if it points to the override
    expect(container.read(provider(0)), isA<MachineStatusIdle<State, Event>>());
    expect(container.read(provider(0).machine), machine2);
  });

  test('overriding the provider overrides provider.state too', () {
    final machine = fooMachine(autostart: false);
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return machine;
    });
    final container = createContainer(
      overrides: [
        provider.overrideWithValue(
            fooMachine(autostart: true, initial: const State.baz())),
      ],
    );
    addTearDown(container.dispose);

    // does not crash
    container.updateOverrides([
      provider.overrideWithValue(machine),
    ]);

    expect(container.read(provider.machine), machine);
    expect(container.read(provider), isA<MachineStatusIdle<State, Event>>());

    machine.start();

    expect(
        container.read(provider),
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.foo()));
  });

  test('can specify name', () {
    final provider = fooMachineProvider(autostart: true, name: 'example');

    expect(provider.machine.name, 'example.machine');
    expect(provider.name, 'example');

    final provider2 = fooMachineProvider(autostart: true);

    expect(provider2.machine.name, isNull);
    expect(provider2.name, isNull);
  });

  test('disposes the notifier when provider is unmounted', () {
    final machine = fooMachine(autostart: false);
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return machine;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    expect(container.read(provider.machine), machine);
    expect(machine.mounted, isTrue);

    container.dispose();

    expect(machine.mounted, isFalse);
  });

  test('provider subscribe the callback is never', () async {
    final machine = fooMachine(autostart: true);
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return machine;
    });
    final listener = ControllerListenerMock();
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider.machine, listener, fireImmediately: true);

    verifyOnly(listener, listener(argThat(isA<Machine<State, Event>>())));

    machine.stop();

    await container.pump();

    verifyNoMoreInteractions(listener);

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('provider subscribe callback never called', () async {
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return fooMachine(autostart: false);
    });
    final listener = ListenerMock();
    final container = createContainer();
    addTearDown(container.dispose);

    container.listen(provider, listener, fireImmediately: true);

    verifyOnly(listener, listener(MachineStatus<State, Event>.idle()));

    container.read(provider.machine).start();

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.running(state: const State.foo())));

    container.dispose();

    verifyNoMoreInteractions(listener);
  });

  test('.notifier obtains the controller without listening to it', () async {
    final dep = StateProvider((ref) => 0);
    final machine = fooMachine(autostart: true);
    final machine2 = fooMachine(autostart: true);
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return ref.watch(dep).state == 0 ? machine : machine2;
    });
    final container = createContainer();
    addTearDown(container.dispose);

    var callCount = 0;
    final sub = container.listen(
      provider.machine,
      (_) => callCount++,
    );

    expect(sub.read(), machine);
    expect(callCount, 0);

    machine.stop();

    await container.pump();

    expect(callCount, 0);

    container.read(dep).state++;

    expect(sub.read(), machine2);

    await container.pump();

    expect(sub.read(), machine2);
    expect(callCount, 1);
  });

  test(
      'overrideWithValue listens to the notifier, support notifier change, and does not dispose of the notifier',
      () async {
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return fooMachine(autostart: false);
    });
    final machine = fooMachine(autostart: true, initial: const State.foo());
    final machine2 = fooMachine(autostart: true, initial: const State.bar());
    final container = createContainer(overrides: [
      provider.overrideWithValue(machine),
    ]);
    addTearDown(container.dispose);
    final listener = ListenerMock();

    container.listen(provider, listener, fireImmediately: true);

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.running(state: const State.foo())));
    expect(container.read(provider.machine), machine);
    // expect(notifier.hasListeners, true);

    machine.stop();

    await container.pump();

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.stopped(lastState: const State.foo())));

    container.updateOverrides([
      provider.overrideWithValue(machine2),
    ]);

    await container.pump();

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.running(state: const State.bar())));

    // expect(machine.hasListeners, false);
    expect(machine.mounted, true);
    expect(container.read(provider.machine), machine2);

    machine2.stop();

    await container.pump();

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.stopped(lastState: const State.bar())));
  });

  test('overrideWithProvider preserves the state accross update', () async {
    final provider = StateMachineProvider<State, Event>.custom((ref) {
      return fooMachine(autostart: false);
    });
    final machine = fooMachine(autostart: true, initial: const State.foo());
    final machine2 = fooMachine(autostart: true, initial: const State.bar());
    final container = createContainer(overrides: [
      provider.overrideWithProvider(
        StateMachineProvider<State, Event>.custom((_) {
          return machine;
        }),
      ),
    ]);
    addTearDown(container.dispose);
    final listener = ListenerMock();

    container.listen(provider, listener, fireImmediately: true);

    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.running(state: const State.foo())));
    expect(container.read(provider.machine), machine);
    // expect(notifier.hasListeners, true);

    machine.stop();

    await container.pump();
    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.stopped(lastState: const State.foo())));

    container.updateOverrides([
      provider.overrideWithProvider(
        StateMachineProvider<State, Event>.custom((_) {
          return machine2;
        }),
      ),
    ]);

    await container.pump();
    expect(container.read(provider.machine), machine);
    // expect(notifier2.hasListeners, false);
    verifyNoMoreInteractions(listener);

    machine.start(stateValue: const State.bar());

    await container.pump();
    expect(container.read(provider.machine), machine);
    verifyOnly(
        listener,
        listener(
            MachineStatus<State, Event>.running(state: const State.bar())));
    expect(machine.mounted, true);

    container.dispose();

    expect(machine.mounted, false);
  });

  group('StateMachineProvider.family', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final machine = fooMachine(autostart: true, initial: const State.foo());
        final provider = StateMachineProvider.family.custom<State, Event, int>(
          (ref, _) => machine,
        );
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider(0).machine), machine);
        expect(container.read(provider(0)),
            MachineStatus<State, Event>.running(state: const State.foo()));
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
          ]),
        );
        expect(root.getAllProviderElementsInOrder(), isEmpty);
      });

      test('when using provider.overrideWithProvider', () async {
        final machine = fooMachine(autostart: true, initial: const State.foo());
        final provider = StateMachineProvider.family.custom<State, Event, int>(
          (ref, _) => machine,
        );
        final root = createContainer();
        final machineOverride =
            fooMachine(autostart: true, initial: const State.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            (value) => StateMachineProvider<State, Event>.custom(
                (ref) => machineOverride),
          ),
        ]);

        expect(container.read(provider(0).machine), machineOverride);
        expect(container.read(provider(0)),
            MachineStatus<State, Event>.running(state: const State.bar()));
        expect(root.getAllProviderElementsInOrder(), isEmpty);
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
          ]),
        );
      });
    });

    test('properly overrides ==', () {
      final family = StateMachineProvider.family.custom<State, Event, int>(
        (ref, _) => fooMachine(autostart: true, initial: const State.foo()),
      );

      expect(family(0), family(0));
      expect(family(1), isNot(family(0)));
      expect(family(1), family(1));
    });
  });

  group('autodispose', () {
    test('can be refreshed', () async {
      var machine = fooMachine(autostart: true, initial: const State.foo());
      final container = createContainer();
      final provider = StateMachineProvider.autoDispose
          .custom<State, Event>((ref) => machine);

      expect(container.read(provider),
          MachineStatus<State, Event>.running(state: const State.foo()));
      expect(container.read(provider.machine), machine);

      machine = fooMachine(autostart: true, initial: const State.bar());
      expect(container.refresh(provider),
          MachineStatus<State, Event>.running(state: const State.bar()));

      expect(container.read(provider),
          MachineStatus<State, Event>.running(state: const State.bar()));
      expect(container.read(provider.machine), machine);
    });

    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final machine = fooMachine(autostart: true, initial: const State.foo());
        final provider = StateMachineProvider.autoDispose
            .custom<State, Event>((ref) => machine);

        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider.machine), machine);
        expect(container.read(provider),
            MachineStatus<State, Event>.running(state: const State.foo()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });

      test('when using provider.overrideWithValue', () async {
        final machine = fooMachine(autostart: true, initial: const State.foo());
        final provider = StateMachineProvider.autoDispose
            .custom<State, Event>((ref) => machine);
        final root = createContainer();
        final machineOverride =
            fooMachine(autostart: true, initial: const State.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithValue(machineOverride),
        ]);

        expect(container.read(provider.machine), machineOverride);
        expect(container.read(provider),
            MachineStatus<State, Event>.running(state: const State.bar()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });

      test('when using provider.overrideWithProvider', () async {
        final machine = fooMachine(autostart: true, initial: const State.foo());
        final provider = StateMachineProvider.autoDispose
            .custom<State, Event>((ref) => machine);
        final root = createContainer();
        final machineOverride =
            fooMachine(autostart: true, initial: const State.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(StateMachineProvider.autoDispose
              .custom<State, Event>((ref) => machineOverride)),
        ]);

        expect(container.read(provider.machine), machineOverride);
        expect(container.read(provider),
            MachineStatus<State, Event>.running(state: const State.bar()));
        expect(root.getAllProviderElements(), isEmpty);
        expect(
          container.getAllProviderElements(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider.machine),
          ]),
        );
      });
    });

    test('StateMachineProviderFamily override', () async {
      final machine2 = fooMachine(autostart: true, initial: const State.bar());

      final provider = StateMachineProvider.autoDisposeFamily
          .custom<State, Event, int>((ref, a) => fooMachine());
      final container = createContainer(
        overrides: [
          provider.overrideWithProvider((a) {
            return StateMachineProvider.autoDispose
                .custom<State, Event>((ref) => machine2);
          }),
        ],
      );
      addTearDown(container.dispose);
      final ownerStateListener = Listener<MachineStatus<State, Event>>();
      final ownerNotifierListener = Listener<Machine<State, Event>>();

      // access in the child container
      // try to read provider.state before provider and see if it points to the override
      final stateSub = container.listen(
        provider(0),
        ownerStateListener,
        fireImmediately: true,
      );

      verify(ownerStateListener(
              MachineStatus<State, Event>.running(state: const State.bar())))
          .called(1);
      verifyNoMoreInteractions(ownerStateListener);

      final notifierSub = container.listen(
        provider(0).machine,
        ownerNotifierListener,
        fireImmediately: true,
      );
      verifyOnly(ownerNotifierListener, ownerNotifierListener(machine2));

      notifierSub.close();

      await container.pump();

      expect(machine2.mounted, true);

      stateSub.close();

      expect(machine2.mounted, true);

      await container.pump();

      await container.pump();

      expect(machine2.mounted, false);
    });

    test('overriding the provider overrides provider.state too', () {
      final provider =
          StateMachineProvider.autoDispose.custom<State, Event>((_) {
        return fooMachine();
      });
      final machine = fooMachine(autostart: true, initial: const State.bar());
      final container = createContainer(
        overrides: [
          provider.overrideWithValue(
              fooMachine(autostart: true, initial: const State.baz())),
        ],
      );
      addTearDown(container.dispose);
      final stateListener = Listener<MachineStatus<State, Event>>();
      final notifierListener = Listener<Machine<State, Event>>();

      // does not crash
      container.updateOverrides([
        provider.overrideWithValue(machine),
      ]);

      container.listen(provider.machine, notifierListener,
          fireImmediately: true);
      verify(notifierListener(machine)).called(1);
      verifyNoMoreInteractions(notifierListener);

      container.listen(provider, stateListener, fireImmediately: true);
      verify(stateListener(
              MachineStatus<State, Event>.running(state: const State.bar())))
          .called(1);
      verifyNoMoreInteractions(stateListener);

      machine.stop();

      verify(stateListener(MachineStatus<State, Event>.stopped(
              lastState: const State.bar())))
          .called(1);
      verifyNoMoreInteractions(notifierListener);
      verifyNoMoreInteractions(stateListener);
    });

    test('can specify name', () {
      final provider = StateMachineProvider.autoDispose.custom<State, Event>(
        (_) => fooMachine(),
        name: 'example',
      );

      expect(provider.machine.name, 'example.machine');
      expect(provider.name, 'example');

      final provider2 = StateMachineProvider.autoDispose
          .custom<State, Event>((_) => fooMachine());

      expect(provider2.machine.name, isNull);
      expect(provider2.name, isNull);
    });

    test('disposes the notifier when provider is unmounted', () {
      final machine = fooMachine();
      final provider = StateMachineProvider.autoDispose.custom<State, Event>(
        (_) => machine,
      );
      final container = createContainer();
      addTearDown(container.dispose);

      container.listen(provider, (value) {});
      expect(machine.mounted, isTrue);

      container.dispose();

      expect(machine.mounted, isFalse);
    });

    test('provider subscribe the callback is never', () async {
      final machine = fooMachine(autostart: true);
      final provider = StateMachineProvider.autoDispose.custom<State, Event>(
        (_) => machine,
      );
      final listener = ControllerListenerMock();
      final container = createContainer();
      addTearDown(container.dispose);

      container.listen(provider.machine, listener, fireImmediately: true);

      verifyOnly(listener, listener(argThat(isA<Machine<State, Event>>())));

      machine.stop();

      await container.pump();

      verifyNoMoreInteractions(listener);

      container.dispose();

      verifyNoMoreInteractions(listener);
    });

    test('provider subscribe callback never called', () async {
      final machine = fooMachine(autostart: true);
      final provider = StateMachineProvider.autoDispose.custom<State, Event>(
        (_) => machine,
      );
      final listener = Listener<MachineStatus<State, Event>>();
      final container = createContainer();
      addTearDown(container.dispose);

      container.listen(provider, listener, fireImmediately: true);

      verifyOnly(
          listener,
          listener(argThat(equals(
              MachineStatus<State, Event>.running(state: const State.foo())))));

      machine.stop();

      verifyOnly(
              listener,
              listener(MachineStatus<State, Event>.stopped(
                  lastState: const State.foo())))
          .called(1);

      container.dispose();

      verifyNoMoreInteractions(listener);
    });

    test('.notifier obtains the controller without listening to it', () async {
      final dep = StateProvider((ref) => 0);
      final machine = fooMachine(autostart: true);
      final machine2 = fooMachine(autostart: true);
      final provider =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return ref.watch(dep).state == 0 ? machine : machine2;
      });
      final container = createContainer();
      addTearDown(container.dispose);

      var callCount = 0;
      final sub = container.listen(provider.machine, (_) => callCount++);

      expect(sub.read(), machine);
      expect(callCount, 0);

      machine.stop();

      expect(callCount, 0);

      container.read(dep).state++;

      expect(sub.read(), machine2);

      await container.pump();

      expect(sub.read(), machine2);
      expect(callCount, 1);
    });

    test('overrideWithProvider preserves the state accross update', () async {
      final provider =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine();
      });
      final machine = fooMachine(autostart: true, initial: const State.foo());
      final machine2 = fooMachine(autostart: true, initial: const State.bar());
      final container = createContainer(overrides: [
        provider.overrideWithProvider(
          StateMachineProvider.autoDispose.custom<State, Event>((_) {
            return machine;
          }),
        ),
      ]);
      addTearDown(container.dispose);
      final listener = Listener<MachineStatus<State, Event>>();

      container.listen(provider, listener, fireImmediately: true);

      verifyOnly(
          listener,
          listener(
              MachineStatus<State, Event>.running(state: const State.foo())));
      expect(container.read(provider.machine), machine);
      // expect(notifier.hasListeners, true);

      machine.stop();

      verifyOnly(
          listener,
          listener(MachineStatus<State, Event>.stopped(
              lastState: const State.foo())));

      container.updateOverrides([
        provider.overrideWithProvider(
          StateMachineProvider.autoDispose.custom<State, Event>((_) {
            return machine2;
          }),
        ),
      ]);

      await container.pump();

      expect(container.read(provider.machine), machine);
      // expect(notifier2.hasListeners, false);
      verifyNoMoreInteractions(listener);

      machine.start(stateValue: const State.baz());

      await container.pump();

      expect(container.read(provider.machine), machine);
      verifyOnly(
          listener,
          listener(
              MachineStatus<State, Event>.running(state: const State.baz())));
      expect(machine.mounted, true);

      container.dispose();

      expect(machine.mounted, false);
    });
  });

  group('StateMachineProvider.autoDisposeFamily', () {
    group('scoping an override overrides all the associated subproviders', () {
      test('when passing the provider itself', () async {
        final machine = fooMachine(autostart: true);

        final provider =
            StateMachineProvider.autoDisposeFamily.custom<State, Event, int>(
          (ref, _) => machine,
        );
        final root = createContainer();
        final container = createContainer(parent: root, overrides: [provider]);

        expect(container.read(provider(0).machine), machine);
        expect(container.read(provider(0)),
            MachineStatus<State, Event>.running(state: const State.foo()));
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
          ]),
        );
        expect(root.getAllProviderElementsInOrder(), isEmpty);
      });

      test('when using provider.overrideWithProvider', () async {
        final machine = fooMachine();
        final provider =
            StateMachineProvider.autoDisposeFamily.custom<State, Event, int>(
          (ref, _) => machine,
        );
        final root = createContainer();
        final machineOverride =
            fooMachine(autostart: true, initial: const State.bar());
        final container = createContainer(parent: root, overrides: [
          provider.overrideWithProvider(
            (value) => StateMachineProvider.autoDispose
                .custom<State, Event>((ref) => machineOverride),
          ),
        ]);

        expect(container.read(provider(0).machine), machineOverride);
        expect(container.read(provider(0)),
            MachineStatus<State, Event>.running(state: const State.bar()));
        expect(root.getAllProviderElementsInOrder(), isEmpty);
        expect(
          container.getAllProviderElementsInOrder(),
          unorderedEquals(<Object?>[
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0)),
            isA<ProviderElementBase>()
                .having((e) => e.origin, 'origin', provider(0).machine),
          ]),
        );
      });
    });

    test('properly overrides ==', () {
      final family = StateMachineProvider.autoDisposeFamily
          .custom<State, Event, int>((ref, _) => fooMachine());

      expect(family(0), family(0));
      expect(family(1), isNot(family(0)));
      expect(family(1), family(1));
    });
  });

  group('XActiveStateRiverpod', () {
    test('nest provider', () async {
      final container = ProviderContainer();
      final deepestChild = StateMachineProvider<State, Event>.custom((ref) {
        return fooMachine(initial: const State.baz());
      });
      final child = StateMachineProvider<State, Event>.custom((ref) {
        return fooMachine(
          initial: const State.bar(),
          onBar: (self) {
            self.nestProvider(ref, deepestChild);
          },
        );
      });
      final parent = StateMachineProvider<State, Event>.custom(
        (ref) {
          return fooMachine(
            autostart: true,
            initial: const State.foo(),
            onBar: (self) {
              self.onEvent<_Next>((event) {
                self.transition(const State.foo());
              });
            },
            onFoo: (self) {
              self.onEvent<_Next>((event) {
                self.transition(const State.bar());
              });
              self.nestProvider(ref, child);
            },
          );
        },
      );

      final subM = container.listen(parent, (value) {});
      addTearDown(subM.close);
      expect(container.read(child), isA<MachineStatusRunning<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusRunning<State, Event>>());

      container.read(parent.machine).send(const Event.next());
      expect(container.read(child), isA<MachineStatusStopped<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusStopped<State, Event>>());

      container.read(parent.machine).send(const Event.next());
      expect(container.read(child), isA<MachineStatusRunning<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusRunning<State, Event>>());

      container.read(parent.machine).stop();
      expect(container.read(child), isA<MachineStatusStopped<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusStopped<State, Event>>());
    });

    test('nest autodispose provider', () async {
      final container = ProviderContainer();
      final deepestChild =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine(initial: const State.baz());
      });
      final child =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine(
          initial: const State.bar(),
          onBar: (self) {
            self.nestAutoDisposeProvider(ref, deepestChild);
          },
        );
      });
      final parent = StateMachineProvider.autoDispose.custom<State, Event>(
        (ref) {
          return fooMachine(
            autostart: true,
            initial: const State.foo(),
            onBar: (self) {
              self.onEvent<_Next>((event) {
                self.transition(const State.foo());
              });
            },
            onFoo: (self) {
              self.onEvent<_Next>((event) {
                self.transition(const State.bar());
              });
              self.nestAutoDisposeProvider(ref, child);
            },
          );
        },
      );

      final subM = container.listen(parent, (value) {});
      addTearDown(subM.close);
      expect(container.read(child), isA<MachineStatusRunning<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusRunning<State, Event>>());

      container.read(parent.machine).send(const Event.next());
      expect(container.read(child), isA<MachineStatusStopped<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusStopped<State, Event>>());

      container.read(parent.machine).send(const Event.next());
      expect(container.read(child), isA<MachineStatusRunning<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusRunning<State, Event>>());

      container.read(parent.machine).stop();
      expect(container.read(child), isA<MachineStatusStopped<State, Event>>());
      expect(container.read(deepestChild),
          isA<MachineStatusStopped<State, Event>>());
    });

    test('executing onState', () {
      final container = ProviderContainer();
      final cb = MockOnStateCb();
      final cbB = MockOnStateCb();
      final child = StateMachineProvider<State, Event>.custom((ref) {
        return fooMachine(
          initial: const State.bar(),
          onBar: (self) {
            self.onEvent<_Next>((event) {
              self.transition(const State.foo());
            });
          },
        );
      });
      final childB =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine(
          initial: const State.baz(),
        );
      });
      final parent =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine(
          autostart: true,
          onFoo: (self) {
            self.nestProvider(ref, child, onState: cb);
            self.nestAutoDisposeProvider(ref, childB, onState: cbB);
          },
        );
      });
      final subM = container.listen(parent, (value) {});
      addTearDown(subM.close);
      container.read(child.machine).send(const Event.next());
      verifyInOrder([
        cb(const State.bar()),
        cbB(const State.baz()),
        cb(const State.foo()),
      ]);
    });

    test('transition in onState', () {
      final container = ProviderContainer();
      final child = StateMachineProvider<State, Event>.custom((ref) {
        return fooMachine(initial: const State.bar());
      });
      final parent =
          StateMachineProvider.autoDispose.custom<State, Event>((ref) {
        return fooMachine(
          autostart: true,
          onFoo: (self) {
            self.nestProvider(ref, child, onState: (s) {
              self.transition(const State.bar());
            });
          },
        );
      });

      final subM = container.listen(parent, (value) {});
      addTearDown(subM.close);
      expect(
          subM.read(),
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.bar()));
    });
  });
}

@freezed
class State with _$State {
  const factory State.foo() = _Foo;
  const factory State.bar() = _Bar;
  const factory State.baz() = _Baz;
}

@freezed
class Event with _$Event {
  const factory Event.next() = _Next;
  const factory Event.toFoo() = _ToFoo;
  const factory Event.toBar() = _ToBar;
  const factory Event.toBaz() = _ToBaz;
}

class ControllerListenerMock extends Mock {
  void call(Machine<State, Event>? value);
}

class OnStateCb extends Mock {
  void call(State state);
}

class ListenerMock extends Mock {
  void call(MachineStatus<State, Event> value);
}

class Listener<T> extends Mock {
  void call(T? value);
}

typedef VerifyOnly = VerificationResult Function<T>(
  Mock mock,
  T matchingInvocations,
);

/// See riverpod test/utils.dart
/// Sorry Remi for copy paste :))) Love your work
///
/// Syntax sugar for:
///
/// ```dart
/// verify(mock()).called(1);
/// verifyNoMoreInteractions(mock);
/// ```
VerifyOnly get verifyOnly {
  final verification = verify;

  return <T>(mock, invocation) {
    final result = verification(invocation);
    result.called(1);
    verifyNoMoreInteractions(mock);
    return result;
  };
}
